
/* Chrysalide - Outil d'analyse de fichiers binaires
 * loop.c - détection des boucles dans du code machine
 *
 * Copyright (C) 2013-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Chrysalide.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "loop.h"



/* Matérialise les liens de retour arrière en tant que boucles. */
static void detect_back_edges(dragon_node *, size_t);



/******************************************************************************
*                                                                             *
*  Paramètres  : nodes = liste de noeuds détectés dans une routine.           *
*                count = taille de cette liste de noeuds à traiter.           *
*                                                                             *
*  Description : Matérialise les liens de retour arrière en tant que boucles. *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void detect_back_edges(dragon_node *nodes, size_t count)
{
    size_t k;                               /* Boucle de parcours #1       */
    dragon_node *node;                      /* Noeud à traiter             */
    const bitfield_t *dominators;           /* Liste de dominateurs        */
    GArchInstruction *last;                 /* Instruction finale de noeud */
    size_t dcount;                          /* Nombre de liens de dest.    */
    size_t i;                               /* Boucle de parcours #2       */
    const instr_link_t *dest;               /* Instr. visée par une autre  */
    dragon_node *target;                    /* Noeud référence à tester    */
    size_t id;                              /* Indice du bit associé       */

    for (k = 0; k < count; k++)
    {
        node = get_dragon_node(nodes, k);

        dominators = get_domination_bits(node);

        get_dragon_node_bounding_instructions(node, NULL, &last);

        g_arch_instruction_lock_dest(last);
        dcount = g_arch_instruction_count_destinations(last);

        for (i = 0; i < dcount; i++)
        {
            dest = g_arch_instruction_get_destination(last, i);

            switch (dest->type)
            {
                case ILT_EXEC_FLOW:
                case ILT_JUMP:
                case ILT_CASE_JUMP:
                case ILT_JUMP_IF_TRUE:
                case ILT_JUMP_IF_FALSE:

                    target = find_node_for_instruction(nodes, count, false, dest->linked);
                    if (target == NULL) break;

                    id = get_dragon_node_index(nodes, target);

                    if (test_in_bit_field(dominators, id))
                        g_arch_instruction_change_link(last, dest->linked, dest->type, ILT_LOOP);

                    break;

                default:
                    break;

            }

            unref_instr_link(dest);

        }

        g_arch_instruction_unlock_dest(last);

    }

}


/******************************************************************************
*                                                                             *
*  Paramètres  : knight = informations quant à la complexité gérée du code.   *
*                                                                             *
*  Description : Détecte les boucles dans du code machine.                    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void detect_loops_in_code(dragon_knight *knight)
{
    dragon_node *nodes;                     /* Liste des noeuds détectés   */
    size_t count;                           /* Taille de cette liste       */

    get_dragon_knight_content(knight, &nodes, &count);

    compute_all_dominators(nodes, count);

    detect_back_edges(nodes, count);

}
